/*
    Copyright (c) 2009 by Joseph Devietti (devietti@cs.washington.edu).

    This file is part of my DMP ASPLOS 2009 Presentation (DAP).

    DAP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DAP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DAP.  If not, see <http://www.gnu.org/licenses/>.
*/ 

﻿package {
	
	import com.boostworthy.animation.sequence.tweens.Action;
	import com.boostworthy.animation.sequence.tweens.Tween;
	import com.dncompute.graphics.*;
	import flash.display.*;
	import flash.geom.*;
	import flash.text.*;
	import com.senocular.drawing.DashedLine;
	import flash.filters.*;
	
	/**
	 * Custom shapes
	 * @author Joe Devietti
	 */
	public class MyShapes {
		
		/** Embed fonts */
		[Embed(systemFont='Tw Cen MT', fontName='fontTwCenMT', mimeType='application/x-font')]
		private static var fontTwCenMT:Class;
		
		public static const INSN_HEIGHT:int = 35;
		public static const INSN_WIDTH:int = 150;
		public static const INSN_YOFFSET:int = INSN_HEIGHT + 5;
		
		public static const MINI_INSN_HEIGHT:int = 25;
		public static const MINI_INSN_WIDTH:int = 100;
		public static const MINI_INSN_YOFFSET:int = MINI_INSN_HEIGHT + 5;
		
		public static const INSN_COLORS:Array =        [ 0x0000CC, 0xFF3333, 0xFFCC33 ];
		
		private static function sub(a:uint,b:uint,mask:uint):uint {
			if ( (a & mask) <= b ) {
				return 0x0;
			} else {
				return (a - b) & mask;
			}
		}
		public static function getDropShadow(color:uint):DropShadowFilter {

			// compute shadow color by subtracting 0x88 from each of RGB - yields a darker version of `color'
			var shadow_color:uint = sub(color,0x880000,0xFF0000) + sub(color,0x008800,0x00FF00) + sub(color,0x000088,0x0000FF);
			
            var angle:Number = 45;
            var alpha:Number = 0.8;
            var blurX:Number = 8;
            var blurY:Number = 8;
            var distance:Number = 5;
            var strength:Number = 0.75;
            var inner:Boolean = false;
            var knockout:Boolean = false;
            var quality:Number = BitmapFilterQuality.HIGH;
            var dsf:DropShadowFilter = new DropShadowFilter(distance, angle, shadow_color, alpha,
				blurX, blurY, strength, quality, inner, knockout);
			return dsf;
		}
		
		/** "Time flows down" arrow */
		public static function TimeArrow():Sprite {
			var style:ArrowStyle = new ArrowStyle();
			style.headLength = 35;
			style.headWidth = 34;
			style.shaftPosition = 0;
			style.shaftThickness = 28;
			style.edgeControlPosition = 0.5;
			style.edgeControlSize = 0.5;
			
			var color:uint = 0x777777;
			
			var arrow:Sprite = new Sprite();
			arrow.graphics.lineStyle( 2, color );
			const length:uint = 120;
			
			GraphicsUtil.drawArrow( arrow.graphics, new Point(0,0), new Point(length,0), style );
			
			// create text
			var tf:TextField = new UnselectableTextField();
			tf.embedFonts = true;
			tf.text = "TIME";
			tf.autoSize = TextFieldAutoSize.CENTER;
			
			// format text
			var format:TextFormat = new TextFormat();
			format.font = "fontTwCenMT";
			format.color = color; //0xFFFFFF;
			format.size = 30;
			format.bold = true;
			tf.setTextFormat( format );
			
			var y_offset:Number = (arrow.height - tf.height) / 2.0;
			
			// DEBUG: draw bounding box
			//arrow.graphics.lineStyle( 1, 0x0 );
			//arrow.graphics.drawRect(0,0,arrow.width,arrow.height);
			
			arrow.addChild( tf );
			tf.x = ((length - style.headLength) - tf.width) / 2.0;
			tf.y = y_offset;
			
			arrow.rotation = 90;
			
			return arrow;
		}
		
		public static function getLabelTextFormat(color:int = 0x0):TextFormat {
			var format:TextFormat = new TextFormat();
			format.font = "Tw Cen MT";
			format.color = color;
			format.size = 30;
			format.bold = false;
			return format;
		}
		private static function getInsnTextFormat(color:int = 0xFFFFFF):TextFormat {
			var format:TextFormat = new TextFormat();
			format.font = "Lucida Console";
			format.color = color;
			format.size = 20;
			format.bold = true;
			return format;
		}
		private static function getMiniInsnTextFormat(color:int = 0xFFFFFF):TextFormat {
			var format:TextFormat = new TextFormat();
			format.font = "Lucida Console";
			format.color = color;
			format.size = 17;
			format.bold = true;
			return format;
		}
		
		/** create a new insn shape */
		public static function Insn(text:String, color:int):Sprite {
			var rect:Sprite = new Sprite();
			rect.graphics.lineStyle();
			rect.graphics.beginFill( color );
			rect.alpha = 1.0;
			// draw the shape
			var height:int = MyShapes.INSN_HEIGHT;
			var width:int = MyShapes.INSN_WIDTH;
			rect.graphics.drawRoundRect( 0, 0, width, height, 15 );
			
			// create a TextField object
			var t:TextField = new UnselectableTextField();
			t.text = text;
			t.autoSize = TextFieldAutoSize.LEFT;
			t.setTextFormat( getInsnTextFormat() );
			t.alpha = 1.0 / rect.alpha;

			// Add the TextField object to the display list
			rect.addChild(t);
			
			// center text within rectangle
			var textbb:Rectangle = t.getBounds( rect );
			// use coords in "rectangle space"
			t.x = ((width - textbb.width) / 2);
			t.y = ((height - textbb.height) / 2);
			
			rect.filters = [ getDropShadow(color) ];
			
			return rect;
		}
		
		/** Create a highlight box that will highlight n insns */
		public static function InsnHighlighter(n:int = 1):Sprite {
			var hilite:Sprite = new Sprite();
			
			// green outline box
			//hilite.graphics.lineStyle( 5, 0x00CC00, 1.0, true, LineScaleMode.NORMAL, CapsStyle.SQUARE, JointStyle.MITER );
			//hilite.graphics.drawRect( 0, 0, MyShapes.INSN_WIDTH, (n*MyShapes.INSN_HEIGHT) );
			
			// background hiliter. tricky to get to work with gray insns boxes
			
			hilite.blendMode = BlendMode.DARKEN;
			hilite.graphics.lineStyle();
			hilite.graphics.beginFill( 0xFFFF00, 1.0 );
			hilite.graphics.drawRect( 0, 0, MyShapes.INSN_WIDTH, (n*MyShapes.INSN_HEIGHT) );
			hilite.graphics.endFill();
			
			
			return hilite;
		}
		
		/** Create a lowlight box that will lowlight n insns */
		public static function InsnLowlighter(n:int = 1):Sprite {
			var lolite:Sprite = new Sprite();
			lolite.graphics.lineStyle();
			lolite.graphics.beginFill( 0x0, 0.75 );
			lolite.graphics.drawRect( 0, 0, MyShapes.INSN_WIDTH, (n * MyShapes.INSN_HEIGHT) );
			lolite.graphics.endFill();
			return lolite;
		}
		
		/** highlight one row of the sharing table */
		public static function ShtabHighlighter():Sprite {
			var hilite:Sprite = new Sprite();
			hilite.blendMode = BlendMode.DARKEN;
			hilite.graphics.lineStyle();
			hilite.graphics.beginFill( 0xFFFF00, 1.0 ); // bright yellow
			hilite.graphics.drawRect( 0, 0, ShtabExec.SHTAB_WIDTH, ShtabExec.SHTAB_ROW_HEIGHT );
			hilite.graphics.endFill();
			return hilite;
		}
		
		/** create a new mini insn shape */
		public static function MiniInsn(text:String, color:int):Sprite {
			var rect:Sprite = new Sprite();
			rect.graphics.lineStyle();
			rect.graphics.beginFill( color );
			rect.alpha = 1.0;
			// draw the shape
			var height:int = MyShapes.MINI_INSN_HEIGHT;
			var width:int = MyShapes.MINI_INSN_WIDTH;
			rect.graphics.drawRoundRect( 0, 0, width, height, 8 );
			
			// create a TextField object
			var t:TextField = new UnselectableTextField();
			t.text = text;
			t.autoSize = TextFieldAutoSize.LEFT;
			t.setTextFormat( getMiniInsnTextFormat() );
			t.alpha = 1.0 / rect.alpha;

			// Add the TextField object to the display list
			rect.addChild(t);
			
			// center text within rectangle
			var textbb:Rectangle = t.getBounds( rect );
			// use coords in "rectangle space"
			t.x = ((width - textbb.width) / 2);
			t.y = ((height - textbb.height) / 2);
			
			rect.filters = [ getDropShadow(color) ];
			
			return rect;
		}
		public static function MiniCodeBox(insns:Array):Sprite {
			var codebox:Sprite = new Sprite();
			codebox.graphics.lineStyle( 1.0, 0x0, 1.0, true, LineScaleMode.NORMAL, CapsStyle.SQUARE, JointStyle.MITER );
			codebox.graphics.beginFill( 0xFFFFFF );
			codebox.graphics.drawRect( 0, 0, 
									   MyShapes.MINI_INSN_WIDTH,
									   (insns.length * MyShapes.MINI_INSN_HEIGHT) );
									
			// add insns to codebox
			var i:uint = 1;
			for each ( var insn:String in insns ) {
				// create a TextField object
				var txt:TextField = new UnselectableTextField();
				txt.text = insn;
				txt.autoSize = TextFieldAutoSize.LEFT;
				txt.setTextFormat( getMiniInsnTextFormat(0x0) );
				
				codebox.addChild( txt );
				
				// center text horizontally within codebox
				var bb:Rectangle = txt.getBounds( codebox );
				// use coords in "rectangle space"
				txt.x = ((codebox.width - bb.width) / 2);
				// balance insns vertically
				var vgap:Number = (codebox.height - (insns.length * txt.height)) / (insns.length+1);
				txt.y = (i * vgap) + ((i-1) * txt.height);
				
				i++;
			}
			
			codebox.filters = [ getDropShadow(0x0) ];
			
			return codebox;
		}
		
		/** Creates a box that can hold n insns */
		public static function CodeBox(insns:Array):Sprite {
			var codebox:Sprite = new Sprite();
			codebox.graphics.lineStyle( 1.0, 0x0, 1.0, true, LineScaleMode.NORMAL, CapsStyle.SQUARE, JointStyle.MITER );
			codebox.graphics.beginFill( 0xFFFFFF );
			codebox.graphics.drawRect( 0, 0, 
									   MyShapes.INSN_WIDTH,
									   (insns.length * MyShapes.INSN_HEIGHT) );
									
			// add insns to codebox
			var i:uint = 1;
			for each ( var insn:String in insns ) {
				// create a TextField object
				var txt:TextField = new UnselectableTextField();
				txt.text = insn;
				txt.autoSize = TextFieldAutoSize.LEFT;
				txt.setTextFormat( getInsnTextFormat(0x0) );
				
				codebox.addChild( txt );
				
				// center text horizontally within codebox
				var bb:Rectangle = txt.getBounds( codebox );
				// use coords in "rectangle space"
				txt.x = ((codebox.width - bb.width) / 2);
				// balance insns vertically
				var vgap:Number = (codebox.height - (insns.length * txt.height)) / (insns.length+1);
				txt.y = (i * vgap) + ((i-1) * txt.height);
				
				i++;
			}
			
			codebox.filters = [ getDropShadow(0x0) ];
			
			return codebox;
		}
		
		public static const TOKEN_RADIUS:uint = 20;
		public static function Token():Sprite {
			
			// draw token shape
			var token:Sprite = new Sprite();
			token.graphics.lineStyle( 2, 0x0, 1.0, true );
			var token_color:uint = 0x33CCFF; // blue
			token.graphics.beginFill( token_color );
			token.graphics.drawCircle( TOKEN_RADIUS, TOKEN_RADIUS, TOKEN_RADIUS ); // first two args specify center point
			token.graphics.endFill();
			
			// add token label
			var txt:TextField = new UnselectableTextField();
			txt.text = "DT";
			txt.autoSize = TextFieldAutoSize.CENTER;
			
			// format text
			var format:TextFormat = new TextFormat();
			format.font = "Lucida Console";
			format.color = 0x0; // black
			format.size = 20;
			format.bold = true;
			txt.setTextFormat( format );

			// Add the TextField object to the display list
			token.addChild( txt );
			
			// center text within circle
			var textbb:Rectangle = txt.getBounds( token );
			// use coords in "circle space"
			txt.x = (((2 * TOKEN_RADIUS) - textbb.width) / 2);
			txt.y = (((2 * TOKEN_RADIUS) - textbb.height) / 2);
			
			token.filters = [ getDropShadow(token_color) ];
			
			return token;
		}
		
		public static const CPU_WIDTH:uint = 40;
		public static function Cpu(label:String):Sprite {
			
			// draw cpu shape
			var cpu:Sprite = new Sprite();
			cpu.graphics.lineStyle( 1, 0x0, 1.0, true );
			var cpu_color:uint = 0x00DD00; // green
			cpu.graphics.beginFill( cpu_color );
			cpu.graphics.drawRoundRect( 0, 0, CPU_WIDTH, CPU_WIDTH, 8 ); // square
			cpu.graphics.endFill();
			
			// add cpu label
			var txt:TextField = new UnselectableTextField();
			txt.text = label;
			txt.autoSize = TextFieldAutoSize.CENTER;
			
			// format text
			var format:TextFormat = new TextFormat();
			format.font = "Lucida Console";
			format.color = 0x0; // black
			format.size = 20;
			format.bold = true;
			txt.setTextFormat( format );

			// Add the TextField object to the display list
			cpu.addChild( txt );
			
			// center text within rectangle
			var textbb:Rectangle = txt.getBounds( cpu );
			// use coords in "rectangle space"
			txt.x = ((CPU_WIDTH - textbb.width) / 2);
			txt.y = ((CPU_WIDTH - textbb.height) / 2);
			
			cpu.filters = [ getDropShadow(cpu_color) ];
			
			return cpu;
		}
		
		public static const CPU_STATUS_PARALLEL_MODE:String = "parallel";
		public static const CPU_STATUS_PARALLEL_MODE_FINISHED:String = "pardone";
		//public static const CPU_STATUS_WAITING_TO_PASS_TOKEN:String = "waiting2pass";
		public static const CPU_STATUS_SERIAL_MODE:String = "serial";
		
		private static const STRING_OF_STATUS:Object = { parallel:"parallel\nmode",
			//waiting2pass:"waiting to\npass token",
			pardone:"parallel\nmode done",
			serial:"serial\nmode" };
			
		public static const CpuStatusTextFormat:TextFormat = new TextFormat("Tw Cen MT",18,0x0,false);
			
		public static function CpuStatus(status:String):Sprite {
			var stat:Sprite = new Sprite();
			
			var txt:TextField = new UnselectableTextField();
			txt.name = "status_text";
			txt.text = STRING_OF_STATUS[status];
			txt.autoSize = TextFieldAutoSize.LEFT;
			
			txt.setTextFormat( CpuStatusTextFormat );
			
			stat.addChild( txt );
			return stat;
		}
		private static function setCpuStatusTextField(tf:TextField,text:String):void {
			tf.text = text;
			tf.setTextFormat( MyShapes.CpuStatusTextFormat );
		}
		/** Cross-fades the change in status text by doing a fadeIncr to 0 alpha, swapping the text via an Action, 
		 * and then fading back to full alpha.
		 * @return an Array of Tween's to add to the timeline.  They collectively take up 2*Anim.FADE frames. */
		public static function updateCpuStatus(sp:Sprite,new_status:String):Array {
			var anims:Array = [];
			
			// TODO: maybe make fades a bit faster...
			var tf:TextField = sp.getChildByName( "status_text" ) as TextField;
			anims.push( ExecLayout.fadeToIncr(tf,0.0) );
			
			// change text with an action
			anims.push( new Action(MyShapes.setCpuStatusTextField, [tf,STRING_OF_STATUS[new_status]], 
				MyShapes.setCpuStatusTextField, [tf,tf.text], Anim.Start()) );
			
			anims.push( ExecLayout.fadeToIncr(tf,1.0) );
			
			return anims;
		}
		
		public static const QUANTUM_WIDTH:uint = 75;
		public static const QUANTUM_HEIGHT:uint = 40;
		private static const ROUNDEDNESS:uint = 20;
		
		private static function HalfQuantum(label:String, y:int, fullheight:int, effheight:int):Sprite {
			var quantum:Sprite = new Sprite();
			quantum.graphics.lineStyle();
			quantum.graphics.beginFill( 0xDDDDDD );
			quantum.graphics.drawRoundRect( 0, y, QUANTUM_WIDTH, fullheight, ROUNDEDNESS );
			quantum.graphics.endFill();
			
			// add cpu label
			var txt:TextField = new UnselectableTextField();
			txt.text = label;
			txt.autoSize = TextFieldAutoSize.CENTER;
			
			// format text
			var format:TextFormat = new TextFormat();
			format.font = "Lucida Console";
			format.color = 0x0; // black
			format.size = 30;
			format.bold = true;
			txt.setTextFormat( format );

			// Add the TextField object to the display list
			quantum.addChild( txt );
			
			// center text within rectangle
			var textbb:Rectangle = txt.getBounds( quantum );
			// use coords in "rectangle space"
			txt.x = ((QUANTUM_WIDTH - textbb.width) / 2);
			txt.y = ((effheight - textbb.height) / 2);
			
			return quantum;
		}
		
		public static function ParallelPrefix():Sprite {
			var quantum:Sprite = MyShapes.HalfQuantum( "P", 0, QUANTUM_HEIGHT+ROUNDEDNESS, QUANTUM_HEIGHT );
			
			// mask out rounded corners on bottom
			var mask:Sprite = new Sprite();
			mask.graphics.lineStyle();
			mask.graphics.beginFill( 0x0 );
			mask.graphics.drawRect( 0, 0, QUANTUM_WIDTH, QUANTUM_HEIGHT );
			mask.graphics.endFill();
			quantum.addChild( mask );
			quantum.mask = mask;
			
			// draw dashed line at bottom
			var dash:DashedLine = new DashedLine( quantum, 4, 4 );
			dash.lineStyle( 3, 0x0, 1.0, true, LineScaleMode.NORMAL, CapsStyle.NONE );
			dash.moveTo( 0, QUANTUM_HEIGHT );
			dash.lineTo( MyShapes.QUANTUM_WIDTH, QUANTUM_HEIGHT );
			
			return quantum;
		}
		
		public static function SerialSuffix():Sprite {
			var quantum:Sprite = MyShapes.HalfQuantum( "S", -ROUNDEDNESS, QUANTUM_HEIGHT+ROUNDEDNESS, QUANTUM_HEIGHT );
			
			// mask out rounded corners on top
			var mask:Sprite = new Sprite();
			mask.graphics.lineStyle();
			mask.graphics.beginFill( 0x0 );
			mask.graphics.drawRect( 0, 0, QUANTUM_WIDTH, QUANTUM_HEIGHT );
			mask.graphics.endFill();
			quantum.addChild( mask );
			quantum.mask = mask;
			
			// draw dashed line at top
			var dash:DashedLine = new DashedLine( quantum, 4, 4 );
			dash.lineStyle( 3, 0x0, 1.0, true, LineScaleMode.NORMAL, CapsStyle.NONE );
			dash.moveTo( 0, 0 );
			dash.lineTo( MyShapes.QUANTUM_WIDTH, 0 );
			
			return quantum;
		}
		
		public static function Quantum():Sprite {
			var quantum:Sprite = MyShapes.HalfQuantum( "Q", 0, QUANTUM_HEIGHT*2, QUANTUM_HEIGHT*2 );
			return quantum;
		}
		
		
			
		public static const XACT_XOFFSET:int = 6;
		public static const XACT_YOFFSET:int = (MyShapes.INSN_YOFFSET - MyShapes.INSN_HEIGHT)+2;
		public static const XACT_WIDTH:int = MyShapes.INSN_WIDTH + (2*MyShapes.XACT_XOFFSET);
		public static const XACT_LINE_THICKNESS:int = 5;
		
		public static function xactHeight(n:Number):int {
			return (n*MyShapes.INSN_YOFFSET) + XACT_YOFFSET;
		}
		
		/** graphic for transaction of n insns */
		public static function Xact(n:Number=1.0):Sprite {
			var xact:Sprite = new Sprite();
			
			var width:int = MyShapes.XACT_WIDTH;
			var height:int = xactHeight( n );
			
			// use new DashedLine each time, to make extending look better (helps a bit)
			var dash:DashedLine = new DashedLine( xact, 3, 3 );
			dash.lineStyle( XACT_LINE_THICKNESS, 0x0, 1.0, true, LineScaleMode.NORMAL, CapsStyle.NONE );
			dash.moveTo( 0, 0 );
			dash.lineTo( 0, height );
			
			dash = new DashedLine( xact, 3, 3 );
			dash.lineStyle( XACT_LINE_THICKNESS, 0x0, 1.0, true, LineScaleMode.NORMAL, CapsStyle.NONE );
			dash.moveTo( width, 0 );
			dash.lineTo( width, height );
			
			dash = new DashedLine( xact, 3, 3 );
			dash.lineStyle( XACT_LINE_THICKNESS, 0x0, 1.0, true, LineScaleMode.NORMAL, CapsStyle.NONE );
			dash.moveTo( 0, 0 );
			dash.lineTo( width, 0 );
			
			// debug commit point placement
			//dash.moveTo( width, height );
			//dash.lineTo( 0, height );
			
			return xact;
		}
		
		public static function XactCommit():Sprite {
			var commit:Sprite = new Sprite();
			var dash:DashedLine = new DashedLine( commit, 3, 3 );
			dash.lineStyle( XACT_LINE_THICKNESS, 0x0, 1.0, true, LineScaleMode.NORMAL, CapsStyle.NONE );

			dash.moveTo( 0, 0 );
			dash.lineTo( MyShapes.XACT_WIDTH, 0 );
			
			return commit;
		}
		
			
		public static function ConflictArrow(x1:int,y1:int,x2:int,y2:int,text:String):Sprite {
			return new TextLineArrow( x1-ExecLayout.COMM_ARROW_XOFFSET, y1, 
				x2+ExecLayout.COMM_ARROW_XOFFSET, y2, text, 0x0, 0xFF0000, 0xFFFFFF );
		}
		
		public static const LABEL_RIGHT:uint = 1;
		public static const LABEL_LEFT:uint = 2;
		
		/** space between object and the label brackets */
		public static const LABEL_OFFSET:int = 10;
		
		/** width of bracket part of label */
		private static const LABEL_BRACKET_WIDTH:int = 5;
		private static const BRACKET_TO_LABEL_GAP:int = 5;
		
		/** 
		 * @param dobj the DisplayObject to be labeled 
		 * @param text a String for the label text
		 * @param position a symbolic constant representing the label position wrt dobj
		 * @param vjitter amount to adjust vertical alignment of label, to compensate for font weirdness */
		public static function Label(dobj:DisplayObject, text:String, position:uint, vjitter:int=0):Sprite {
			
			var label_color:uint = 0x0; // black
			
			// create text
			var tf:TextField = new UnselectableTextField();
			tf.embedFonts = true;
			tf.text = text;
			tf.autoSize = TextFieldAutoSize.CENTER;
			
			// format text
			var format:TextFormat = new TextFormat();
			format.font = "fontTwCenMT";
			format.color = label_color;
			format.size = 30;
			format.bold = false;
			tf.setTextFormat( format );
			
			var label:Sprite = new Sprite();
			label.graphics.lineStyle( 3, label_color, 1.0, true, LineScaleMode.NORMAL, CapsStyle.ROUND, JointStyle.ROUND );
			label.graphics.moveTo( 0, 0 );
			
			switch (position) {
			case LABEL_RIGHT:
				// draw bracket 		TODO: draw a curly brace instead
				label.graphics.lineTo( LABEL_BRACKET_WIDTH, 0 );
				label.graphics.lineTo( LABEL_BRACKET_WIDTH, dobj.height );
				label.graphics.lineTo( 0, dobj.height );
				
				label.addChild( tf );
				tf.x = LABEL_BRACKET_WIDTH + BRACKET_TO_LABEL_GAP;
				// NB: this is v-centered based on the height of capital letters :-(
				tf.y = ((dobj.height - tf.height) / 2.0) + vjitter;
				break;
			case LABEL_LEFT:
				// draw bracket 		TODO: draw a curly brace instead
				label.graphics.lineTo( -LABEL_BRACKET_WIDTH, 0 );
				label.graphics.lineTo( -LABEL_BRACKET_WIDTH, dobj.height );
				label.graphics.lineTo( 0, dobj.height );
				
				label.addChild( tf );
				tf.x = -LABEL_BRACKET_WIDTH - BRACKET_TO_LABEL_GAP - (tf.width);
				
				//var bounds:Rectangle = tf.getBounds(tf);
				//trace( [ bounds.height, tf.height, tf.textHeight, tf.getLineMetrics(0).height, tf.getCharBoundaries(0).height ] );
				// NB: this is v-centered based on the height of capital letters - none of the above help :-(
				
				tf.y = ((dobj.height - tf.height) / 2.0) + vjitter;
				break;
			default:
				throw new Error( "Invalid position specified: " + position );
			}
			
			return label;
		}
		
	} // end class MyShapes
	
}