/*
    Copyright (c) 2009 by Joseph Devietti (devietti@cs.washington.edu).

    This file is part of my DMP ASPLOS 2009 Presentation (DAP).

    DAP is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DAP is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DAP.  If not, see <http://www.gnu.org/licenses/>.
*/ 

﻿package  {
	
	import com.boostworthy.events.AnimationEvent;
	import com.boostworthy.animation.rendering.RenderMethod;
	import flash.display.Sprite;
	import flash.display.Stage;
	import flash.events.*;
	import flash.ui.Mouse;
	import flash.ui.Keyboard;
	import flash.utils.Timer;
	
	import com.boostworthy.animation.sequence.Timeline;
	import com.boostworthy.animation.sequence.tweens.Action;
	import com.boostworthy.core.Global;
	
	import flash.external.ExternalInterface;
	
	/**
	 * Scene manager: builds a scene abstraction on top of the timeline, with scene
	 * transitions triggered by user input events (eg clicks).
	 * @author Joe Devietti
	 */
	public class SceneManager {
		
		private static var g_timeline:Timeline = null;
		private static var g_animateFun:Function = null;
		
		public static const SCENE0_START:uint = 2;
		public static const SCENE0_END:uint = 3;
		private static const INITIAL_SCENES:Array = [SCENE0_START, SCENE0_END, /*other scenes...*/ ];
		private static var g_scene_frames:Array = INITIAL_SCENES;
		
		private static var g_playing_forward:Boolean = true;
		
		/** List of teardown Function's to be called when timeline rolls over. 
		 * Stored in order in which they should be called, ie last scene first.  */
		private static var g_teardown_funs:Array = [];
		
		
		private static function resetAnimation():void {
			for each ( var teardown:Function in g_teardown_funs ) {
				teardown.call( null );
			}			
			//trace( "global.stage's only child is named: " + Global.stage.getChildAt(0).name );
			
			// clear canvas to prevent any corruption at very beginning of next anim
			//trace( "stage has # children: " + Global.stage.numChildren );
			var canvas:Sprite = Global.stage.getChildAt(0) as Sprite;
			for ( var i:int = 0; i < canvas.numChildren; i++ ) {
				canvas.removeChildAt( 0 );
			}
			canvas.graphics.clear();
			canvas.visible = false;
		}
		
		/** input handlers */
		private static function pause():void {
			trace( "paused" );
			g_timeline.stop();
		}
		private static function goahead(event:MouseEvent):void {
			if ( SceneManager.isFinished ) {
				trace( "cleaning up..." );

				SceneManager.isFinished = false;
				
				resetAnimation();

				timer = new Timer( 100, 1 ); // 1-shot timer, 100 ms from now
				timer.addEventListener( TimerEvent.TIMER_COMPLETE, 
					function timerHandler(e:TimerEvent):void{
						trace( "going to next slide..." );
						if ( ExternalInterface.available ) {
							ExternalInterface.call( "gotoNextSlide" );
						}
					});
				timer.start();
				
				return;
			}

			if ( SceneManager.isPlaying ) {
				// skip ahead to end of this scene/start of next scene
				g_timeline.stop();
				var scenes:Array = g_scene_frames.filter( function(fn:int, idx:int, arr:Array):Boolean { return fn > g_timeline.getCurrentFrame(); } );
				g_timeline.gotoAndPlay( scenes.shift() );
				
			} else {
				g_timeline.play();
			}
		}
		private static function keypressed(event:KeyboardEvent):void {

			// HACK: super-fragile to hard-code values here, but Keyboard class doesn't 
			// have the right constants for some reason...
			
			// step animation backwards
			if ( event.keyCode == 66 || event.keyCode == 67 || event.keyCode == 68 ) { // b/c/d keys
				if ( SceneManager.isFinished ) {
					SceneManager.isFinished = false;
				}
				if ( SceneManager.isPlaying ) {
					// skip back to beginning of this scene/end of previous scene
					g_timeline.stop();
					var scenes:Array = g_scene_frames.filter( function(fn:int, idx:int, arr:Array):Boolean { return fn < g_timeline.getCurrentFrame(); } );
					g_timeline.gotoAndPlayReverse( scenes.pop() );

				} else {
					g_timeline.playReverse();
				}
			}
			
			// reset animation to beginning
			if ( event.keyCode == 82 ) { // R key
				/*
				trace( "resetting..." );
				resetAnimation();
				g_timeline.init( RenderMethod.ENTER_FRAME, 30 );
				isPlaying = isFinished = false;
				g_playing_forward = true;
				g_teardown_funs = [];
				g_scene_frames = INITIAL_SCENES;
				Anim.setCurrentAnimationFrame( 0 );
				
				setup(g_animateFun, g_timeline, Global.stage );
				
				trace( "relaunching animation..." );
				*/
			}
		}
		private static function nop():void { }
		
		/** animation state listeners */
		private static var isPlaying:Boolean = false;
		private static var isFinished:Boolean = false;
		private static var timer:Timer = null;
		
		private static function anim_started(e:AnimationEvent):void {
			SceneManager.isPlaying = true;
			//trace( SceneManager.isPlaying );
		}
		private static function anim_stopped(e:AnimationEvent):void {
			SceneManager.isPlaying = false;
			//trace( SceneManager.isPlaying );			
		}
		private static function anim_finished(e:AnimationEvent):void {
			trace( "finished!" );
			SceneManager.isFinished = true;
		}

		/** Returns the current Timeline object */
		public static function TL():Timeline {
			return g_timeline;
		}
		
		/** Initialize the SceneManager.  Must be called before any other method. 
		 * @param timeline the Animation System timeline being used
		 * @param start_frames an array holding the start frames of each scene 
		 * (stop frames are inferred from start frames) */
		public static function setup(/*animate:Function,*/ timeline:Timeline, stage:Stage):void {
			
			//g_animateFun = animate;
			g_timeline = timeline;
			
			// need to stop timeline immediately to prevent addTween()'s re-rendering
			// of frames from resulting in Action's being dispatched multiple times
			g_timeline.gotoAndStop( 1 );

			// setup input handlers
			Global.stage = stage;
			stage.addEventListener( MouseEvent.CLICK, goahead );
			// TODO: doesn't work - only gets registered as a single click...:-(
			stage.addEventListener( MouseEvent.DOUBLE_CLICK, keypressed );
			stage.addEventListener( KeyboardEvent.KEY_DOWN, keypressed );
			
			// works in ActiveX containers, too!
			Mouse.hide();
			// TODO: try to move mouse to trigger hiding...
			
			// setup timeline state listeners
			g_timeline.addEventListener( AnimationEvent.START, anim_started );
			g_timeline.addEventListener( AnimationEvent.STOP, anim_stopped );
			g_timeline.addEventListener( AnimationEvent.FINISH, anim_finished );
			
			// pause at end of scene 0
			g_timeline.addTween( new Action(SceneManager.pause, [], SceneManager.pause, [], SCENE0_END) );
			
			
			//trace( "anim starts @ " + Anim.Start() );
			
			// TODO: BELOW CODE FOR RESETTING - doesn't work yet
			
			// build animations
			//g_animateFun();
			
			// extend timeline until last frame
			//g_timeline.addTween( new Action(SceneManager.nop, [], SceneManager.nop, [], g_scene_frames[g_scene_frames.length-1]) );
			// call teardown funs in *reverse* order when we rollover
			//g_teardown_funs.reverse();
			// get this party started
			//g_timeline.play();
		}
		
		public static function noMoreScenes():void {
			// extend timeline until last frame
			g_timeline.addTween( new Action(SceneManager.nop, [], SceneManager.nop, [], g_scene_frames[g_scene_frames.length-1]) );
			
			// call teardown funs in *reverse* order when we rollover
			g_teardown_funs.reverse();
			
			// start the ball rolling
			g_timeline.play();
			
			// see if this avoid having to move mouse to get it hidden again (it doesn't :-()
			Mouse.hide();
		}
		
		/** Add a new scene to the timeline.
		 * @setup function(start_frame:uint,end_frame:uint):void called at beginning of scene. 
		 * Should setup all sprites+animations for the scene.
		 * @teardown function():void called when scene is played in reverse. Should undo actions of setup function.
		 * @duration the number of frames the scene should last */
		public static function addScene( setup:Function, teardown:Function, duration:uint ):void {
			if ( null == setup || null == teardown ) {
				throw new Error( "setup/teardown functions must be non-null!" );
			}
			if ( duration < 2 ) {
				throw new Error( "duration must be > 1" );
			}
			
			var start_frame:int = g_scene_frames[ g_scene_frames.length-1 ] + 1;
			var end_frame:int = start_frame + duration + 2; // +2 is fudge due to doscene() timing nonsense below
			g_scene_frames.push( start_frame );
			g_scene_frames.push( end_frame );
			
			// setup scene
			var act:Action = new Action(SceneManager.doscene, [setup,end_frame], teardown, [], start_frame);
			g_timeline.addTween( act );
			
			// register teardown fun
			g_teardown_funs.push( teardown );
			
			// pause at end of scene
			g_timeline.addTween( new Action(SceneManager.pause, [], SceneManager.pause, [], end_frame) );
		}
		
		/** Add a new scene that generates a powerpoint "next animation" action (ie, click) going forward, 
		 * and a "reverse animation" action when played in reverse. */
		public static function addPowerpointClick():void {
			SceneManager.addScene( function(start_frame:uint, end_frame:uint):void {
					if ( !ExternalInterface.available ) {
						//throw new Error( "No external interface!" );
						trace( "NO EXTERNAL INTERFACE!" );
					} else {
						g_timeline.addTween( new Action(ExternalInterface.call, ["gotoNextSlide"], 
							ExternalInterface.call, ["gotoPrevSlide"], start_frame) );
					}
					Anim.StopIncr(3);
				},
				function():void {}/*nothing to teardown*/, 3);
/* VBA code to be inserted for each slide+flash animation combination
Private Sub ShockwaveFlash1_FlashCall(ByVal request As String)
    'request looks like:
    '<invoke name="foo" returntype="xml"><arguments></arguments></invoke>
    If InStr(request, "gotoNextSlide") > 0 Then
        ActivePresentation.SlideShowWindow.View.Next
    ElseIf InStr(request, "gotoPrevSlide") > 0 Then
        ActivePresentation.SlideShowWindow.View.Previous
    End If
End Sub
*/
		}
		
		/** Launch a scene by stopping the timeline, running scene setup, and 
		 * then restarting the timeline again.  We have to pause things to ensure that 
		 * we don't get into an infinite loop, as each addTween() causes the current 
		 * frame to be re-rendered, which triggers any actions attached to the frame... */
		public static function doscene(setup:Function, end_frame:uint):void {
			var current_frame:uint = g_timeline.getCurrentFrame();
			
			//trace( "current frame: "+current_frame );
			
			Anim.setCurrentAnimationFrame( current_frame + 2 );
			
			g_timeline.gotoAndStop( current_frame + 1 ); // stop on an empty frame
			// Apply animations starting with next frame. 
			// This fun will cause g_current_animation_frame to be incremented.
			setup.apply( null, [current_frame + 2, end_frame] );
			
			// check that animation timings are sane
			if ( Anim.getCurrentAnimationFrame() != end_frame ) {
				throw new Error( "Animation should have ended at frame "+end_frame+" but instead ended at frame "+Anim.getCurrentAnimationFrame() );
			}
			
			g_timeline.gotoAndPlay( current_frame + 2 );
		}
		
	}
	
}